"""
Utility methods using for files and archives manipulation and filtering.
"""
import fnmatch
import logging
import os

from .constants import APP_NAME, FILE_MODE_FROM_URL

logger = logging.getLogger(APP_NAME)


def match_filters(file, patterns):
    for p in patterns:
        if fnmatch.fnmatchcase(file, p):
            return True
    return False


def filter_files(files, patterns):
    filtered_files = []

    for f in files:
        if match_filters(f, patterns):
            filtered_files.append(f)
    return filtered_files


def get_filtered_files(files, patterns, tar):
    root_dirs = []

    filtered_objects = set([])
    for f in files:
        if match_filters(f, patterns):
            fi = tar.getmember(f)
            if fi.isfile():
                filtered_objects.add(f)
            elif fi.isdir():
                root_dirs.append(f)

    root_dirs.sort(key=len)
    for d in root_dirs:
        pattern = "{}{}*".format(d.split(os.sep)[-1],
                                 os.sep)
        matched_subfiles = filter_files(files, [pattern])
        if len(matched_subfiles) == 0:
            if len(d.split(os.sep)) > 1:  # empty subdirectory
                filtered_objects.add(d)
        else:
            filtered_objects.update(matched_subfiles)
            for subfile in matched_subfiles:
                if subfile in root_dirs:
                    root_dirs.remove(subfile)
    return filtered_objects


def change_tarinfo_properties(tarinfo):
    """
    Filter for insertion local file to tar archive.
    :param tarinfo: tarinfo to insert
    :return: original tarinfo with changed permissions
    """
    tarinfo.uid = 0
    tarinfo.gid = 0
    return tarinfo


def change_tarinfo_properties_mode(tarinfo):
    """
    Filter for insertion not-local file to tar archive.
    :param tarinfo: tarinfo to insert
    :return: original tarinfo with changed permissions
    """
    tarinfo.uid = 0
    tarinfo.gid = 0
    tarinfo.mode = FILE_MODE_FROM_URL
    return tarinfo


def normalise_patterns(patterns):
    normpattern = []
    for p in patterns:
        normpattern.append(os.path.normpath(p))
    return normpattern


def remove_top_parent(path):
    return os.sep.join(path.split(os.sep)[1:])


def is_top_file(path):
    return len(path.split(os.sep)) > 1


def put_archive(client, config, container, tar, target):
    logger.debug('------> creating target folder:\n{}'.format(target))
    create_directory(client, config, container, target)
    client.put_archive(container=container,
                       path=target,
                       data=tar.getvalue())


def create_directory(client, config, container, target):
    client.execute(container=container,
                   cmd=config.mkdir_command + ' {}'.format(
                       target))


def copy_files_to_container(container, client, config, context, source, target):
    single_file = None
    if target != os.curdir and target[-1] != os.path.sep:  # destination is file
        single_file = os.path.basename(target)
        target = os.path.dirname(os.path.join(os.curdir, target))

    tar = context.get_archive(patterns=source,
                              single_file=single_file)

    if not tar:
        logger.warning('------> nothing to copy\n{}'.format(target))
        return

    put_archive(client, config, container, tar, target)


def norm_path_with_work_dir(path, workdir):
    if not workdir:
        return path

    has_sep = path[-1] == os.sep
    return_path = os.path.normpath(os.path.join(workdir, path))
    if has_sep:
        return_path += os.sep
    return return_path



import logging

from .constants import APP_NAME

logger = logging.getLogger(APP_NAME)


class ImageCache:
    def __init__(self):
        pass

    def add(self, image_id, parent_id, layer):
        logger.info('======> CACHE - SAVE\n')
        logger.debug('id: {}\nparent: {}\n{}\n'.format(str(image_id), str(parent_id), str(layer)))

    def get_layer(self, parent, layer):
        logger.info('======> CACHE - IMAGE NOT FOUND\n')
        logger.debug('parent: {}\n{}\n'.format(str(parent), str(layer)))
        return None



"""
Utility methods for builder.
"""
import io
import logging
import os
import tarfile
import tempfile

import six

from .constants import (APP_NAME, COMMIT_MESSAGE_START, CONTEXT_FILE_IN_MEMORY,
                        CONTEXT_FILE_ON_DISK, DEFAULT_CONTEXT_FILE_LIMIT)

logger = logging.getLogger(APP_NAME)


def dockerfile_from_context(context, path):
    with tarfile.open(mode="r", fileobj=context) as t:
        df = t.extractfile(t.getmember(path))
    dockerfile = six.BytesIO()
    dockerfile.writelines(df.readlines())
    dockerfile.seek(0)
    context.seek(0)
    return dockerfile


def mkbuildcontext(path, fileobj, custom_context, dockerfile, client, limit=None):
    if path is None and fileobj is None:
        raise TypeError("Either path or fileobj needs to be provided.")

    dockerfile = dockerfile or "Dockerfile"

    if custom_context:
        if not fileobj:
            raise TypeError("You must specify fileobj with custom_context")
        return fileobj, dockerfile_from_context(fileobj, dockerfile)
    elif fileobj is not None:
        return make_build_context_from_dockerfile(fileobj), fileobj
    elif path.startswith(('http://', 'https://',
                          'git://', 'github.com/', 'git@')):
        # TODO
        return None, None
    elif not os.path.isdir(path):
        raise TypeError("You must specify a directory to build in path")
    else:
        dockerignore = os.path.join(path, '.dockerignore')
        exclude = None
        if os.path.exists(dockerignore):
            with open(dockerignore, 'r') as f:
                exclude = list(filter(bool, f.read().splitlines()))
        context = tar(
            path, client, exclude=exclude, dockerfile=dockerfile, limit=limit
        )
        dockerfile_obj = dockerfile_from_context(context=context, path=dockerfile)
        return context, dockerfile_obj


def exclude_paths(root, exclude, dockerfile, client):
    files = client.exclude_paths_from_context(root, exclude, dockerfile)
    return files


def tar(path, client, exclude=None, dockerfile=None, fileobj=None, limit=DEFAULT_CONTEXT_FILE_LIMIT):
    if not fileobj:
        fileobj = tempfile.SpooledTemporaryFile(max_size=limit)

    root = os.path.abspath(path)
    exclude = exclude or []

    logger.debug("------> context:")
    with tarfile.open(mode='w', fileobj=fileobj) as t:
        for p in sorted(exclude_paths(root, exclude,
                                      dockerfile=dockerfile,
                                      client=client)):
            logger.debug("          - {}".format(p))
            t.add(name=os.path.join(path, p), arcname=p)

    fileobj.seek(0)

    if isinstance(fileobj, tempfile.SpooledTemporaryFile):
        if fileobj.name:
            context_file_type = CONTEXT_FILE_ON_DISK
        else:
            context_file_type = CONTEXT_FILE_IN_MEMORY
        logger.debug("------> context type: {}".format(context_file_type))

    return fileobj


def make_build_context_from_dockerfile(dockerfile):
    f = six.BytesIO()
    with tarfile.open(mode='w', fileobj=f) as t:
        if isinstance(dockerfile, six.StringIO) or isinstance(dockerfile, io.StringIO):
            dfinfo = tarfile.TarInfo('Dockerfile')
            dfinfo.size = len(dockerfile.getvalue())
            dockerfile.seek(0)
        elif isinstance(dockerfile, six.BytesIO) or isinstance(dockerfile, io.BytesIO):
            dfinfo = tarfile.TarInfo('Dockerfile')
            dfinfo.size = len(dockerfile.getvalue())
            dockerfile.seek(0)
        else:
            dfinfo = t.gettarinfo(fileobj=dockerfile, arcname='Dockerfile')
        t.addfile(dfinfo, dockerfile)
    f.seek(0)
    return f


def get_name_and_tag(name):
    name_parts = name.split(':')
    if len(name_parts) == 2:
        repo = name_parts[0]
        tag = name_parts[1]
    else:
        repo = name
        tag = "latest"

    return repo, tag


class ImageLayer(object):
    def __init__(self, number, commands=None):
        self._commands = commands or []
        self.number = number
        self.is_last = False
        self.commit_message = COMMIT_MESSAGE_START
        self.conf = LayerConfig()

    def __str__(self):
        commands = "["
        for cmd in self.commands:
            commands += "<{}>".format(str(cmd))
        commands += "]"
        return "layer {}: {}".format(str(self.number), commands)

    def add_command(self, command, content):
        self._commands.append(command)

        formated_content = " {}".format(content.rstrip())
        self.commit_message += formated_content
        label_name = "layer_{}_commands".format(str(self.number))
        self.conf.commit_config.setdefault('Labels', {})
        self.conf.commit_config['Labels'].setdefault(label_name, "")
        self.conf.commit_config['Labels'][label_name] += formated_content

    @property
    def commands(self):
        return self._commands


class ResultImage(object):
    def __init__(self, image_id, logs=None):
        self._id = image_id
        self._logs = logs or []

    @property
    def id(self):
        return self._id

    @property
    def logs(self):
        return self._logs

    def add_log(self, log):
        self._logs.append(log)

    def __str__(self):
        return "ResultImage<id: {}>".format(str(self._id))


class CommandLog(object):
    def __init__(self, command, logs=None):
        self._command = command
        self._logs = logs or []

    @property
    def command(self):
        return self._command

    @property
    def logs(self):
        return self._logs

    def add_log(self, log):
        self._logs.append(log)

    def add_logs(self, logs):
        self._logs += logs


class LayerConfig(object):
    def __init__(self):
        self.commit_config = {}
        self.container = None
        self.author = ""
        self._logs = []
        self._workdir = os.curdir

    @property
    def workdir(self):
        return self._workdir

    @workdir.setter
    def workdir(self, workdir):
        new_workdir = os.path.normpath(os.path.join(self._workdir, workdir))
        self.add_property("WorkingDir", new_workdir)
        self._workdir = new_workdir

    def add_property(self, key, value):
        self.commit_config[key] = value

    def add_labels(self, labels):
        self.commit_config.setdefault("Labels", {})
        self.commit_config["Labels"].update(labels)

    def add_envs(self, envs):
        self.commit_config.setdefault("Env", [])
        for k, v in envs.items():
            self.commit_config["Env"].append('{}:{}'.format(str(k), str(v)))
        unique_envs = list(set(self.commit_config["Env"]))
        self.commit_config["Env"] = unique_envs

    @property
    def logs(self):
        return self._logs

    def add_log(self, log):
        self._logs.append(log)


def volumes_to_list(dict, mode=""):
    if mode:
        mode = ":" + mode
    else:
        mode = ""

    volumes = []
    for key, value in six.iteritems(dict):
        volumes.append("{}:{}{}".format(key, value, mode))
    return volumes


def volumes_to_bind_dict(dict, mode=""):
    volumes = {}
    for key, value in six.iteritems(dict):
        volumes[key] = {"bind": value,
                        "mode": mode
                        }
    return volumes


def volumes_to_dict(volumes):
    volumes = get_list_from_tuple_or_string(volumes)
    result = {}

    for v in volumes:
        v_split = v.split(':')
        result[v_split[0]] = v_split[1]

    return result


def get_list_from_tuple_or_string(value):
    if not value:
        return []
    if isinstance(value, six.string_types):
        return [value]
    else:
        return list(value)



"""
Class for manipulating with build context.
"""
import os
import tarfile

import six

from .fileutils import (get_filtered_files, is_top_file, normalise_patterns,
                        remove_top_parent)


class BuildContext:
    def __init__(self, context, limit):
        self._context = context
        self._struct = None
        self._context.seek(0)
        self.limit = limit

    @property
    def structure(self):
        if not self._struct:
            self._context.seek(0)
            with tarfile.open(mode="r", fileobj=self._context) as tar_file:
                self._struct = tar_file.getnames()
            self._context.seek(0)
        return self._struct

    def get_file(self, name):
        try:
            with tarfile.open(mode="r", fileobj=self._context) as tar:
                f = tar.extractfile(tar.getmember(name))
            if not f:
                return None
            file = six.BytesIO()
            file.writelines(f.readlines())
            file.seek(0)
            return file
        except:
            return None
        finally:
            self._context.seek(0)

    def get_archive(self, patterns, single_file=None):

        normpatterns = normalise_patterns(patterns)

        with tarfile.open(mode="r", fileobj=self._context) as tar_file:

            put_all = os.curdir in normpatterns

            files_to_put = self.structure if put_all else get_filtered_files(self.structure, normpatterns, tar_file)

            if len(files_to_put) == 0:
                return None
            elif single_file and len(files_to_put) > 1:
                raise ValueError("expecting single file in patterns: {}".format(normpatterns))

            tar_in_memory = six.BytesIO()
            with tarfile.open(mode="w", fileobj=tar_in_memory) as filtered_tar:
                for file in files_to_put:
                    file_info = tar_file.getmember(file)
                    file_object = tar_file.extractfile(file_info)
                    if single_file:
                        file_info.name = single_file
                    elif not put_all and is_top_file(file):
                        file_info.name = remove_top_parent(file)
                    filtered_tar.addfile(tarinfo=file_info,
                                         fileobj=file_object)

            tar_in_memory.seek(0)
            self._context.seek(0)
            return tar_in_memory



"""
Constants used by builder.
"""

CONFIG_LAYERS = "layers"
CONFIG_VOLUMES = "volumes"
CONFIG_LABELS = "labels"
CONFIG_VERSION = "version"
CONFIG_INFINITE_COMMAND = "infinite_command"
CONFIG_REPOSITORY = "repository"
CONFIG_TAGS = "tags"
CONFIG_MKDIR_COMMAND = 'mkdir_command'
CONFIG_BUILDARGS = 'buildargs'
CONFIG_CONTAINER_LIMITS = 'container_limits'
CONFIG_PULL = "pull"

DEFAULT_REPOSITORY = "incubator-image"
DEFAULT_TAG = "latest"

INFINITE_COMMAND = "cat -"
MKDIR_COMMAND = 'mkdir --parent'

DEFAULT_VERSION = 1
COMMIT_MESSAGE_START = "This layer is composed of these commands:"

FILE_MODE_FROM_URL = 0o600

APP_NAME = 'incubator'

CONFIG_CONTEXT_FILE_LIMIT = "context_file_limit"
DEFAULT_CONTEXT_FILE_LIMIT = 0
CONTEXT_FILE_IN_MEMORY = "in memory"
CONTEXT_FILE_ON_DISK = "on disk"

MAX_API_ATTEMPTS = 10



import json
import logging

import click

import incubator

from .. import __version__ as VERSION
from ..core.constants import APP_NAME
from ..core.utils import get_list_from_tuple_or_string


def print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo("{}-{}".format(APP_NAME, VERSION))
    ctx.exit()


CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
@click.option('--version', '-V', is_flag=True, callback=print_version,
              expose_value=False, is_eager=True)
def cli():
    """
    Incubator is an alternative image builder for docker containers.
    """
    pass


@click.command(context_settings=CONTEXT_SETTINGS)
@click.argument('path', type=click.Path(exists=True,
                                        file_okay=False,
                                        dir_okay=True,
                                        readable=True))
@click.option('--build-arg', multiple=True, type=click.STRING, help="Set build-time variables (default [])")
@click.option('--cpu-shares', '-c', type=click.INT, help="CPU shares (relative weight).")
@click.option('--config', '-g', type=click.File(), help="File with config.")
@click.option('--context-file-limit', type=click.INT, help="Limit for in-memory context. Default is 0=unlimited.")
@click.option('--cpuset-cpus', type=click.STRING, help="CPUs in which to allow execution (0-3, 0,1)")
@click.option('--file', '-f', type=click.Path(exists=False), help="Name of the Dockerfile.")
@click.option('--force_rm', is_flag=True, help="Always remove intermediate containers")
@click.option('--label', '-l', multiple=True, type=click.STRING, help="Set metadata for an image (default [])")
@click.option('--memory', '-m', type=click.STRING, help="Memory limit.")
@click.option('--memory-swap', type=click.STRING,
              help="Swap limit equal to memory plus swap: '-1' to enable unlimited swap.")
@click.option('--no-cache', is_flag=True, help=" Do not use cache when building the image.")
@click.option('--pull', is_flag=True, help="Always attempt to pull a newer version of the image")
@click.option('--quiet', '-q', is_flag=True, help="Only display ID.")
@click.option('--rm', is_flag=True, help="Remove intermediate containers after a successful build (default true)")
@click.option('--tag', '-t', multiple=True, type=click.STRING, help="Tag for final image.")
@click.option('--test-config', is_flag=True, help="Print only given configuration and exit.")
@click.option('--verbose', is_flag=True, help="Be verbose.")
@click.option('--volume', '-v', type=click.STRING, help="Set build-time bind mounts (default [])")
def build(path, build_arg, cpu_shares, config,
          context_file_limit,
          cpuset_cpus, file, force_rm,
          label, memory, memory_swap,
          no_cache, pull, quiet, rm,
          tag, test_config, verbose, volume):
    """
    Alternative to docker build command.

    Build process can be controlled  with config file.
    """

    if config:
        config_dict = json.loads(config.read())
    else:
        config_dict = {}

    if verbose:
        incubator.set_logging(level=logging.DEBUG, add_handler=(not quiet))
    else:
        incubator.set_logging(level=logging.INFO, add_handler=(not quiet))

    volume = get_list_from_tuple_or_string(value=volume)
    tag = get_list_from_tuple_or_string(value=tag)
    label = get_list_from_tuple_or_string(value=label)

    limits = {
        'memory': memory,
        'memswap': memory_swap,
        'cpushares': cpu_shares,
        'cpusetcpus': cpuset_cpus
    }

    if test_config:
        click.echo("path: {}\n"
                   "build_args: {}\n"
                   "cpu_shares: {}\n"
                   "config: {}\n"
                   "context_file_limit: {}\n"
                   "cpuset_cpus: {}\n"
                   "file: {}\n"
                   "force_rm: {}\n"
                   "labels: {}\n"
                   "memory: {}\n"
                   "memory_swap: {}\n"
                   "no_cache: {}\n"
                   "pull: {}\n"
                   "quiet: {}\n"
                   "rm: {}\n"
                   "tags: {}\n"
                   "test_config: {}\n"
                   "verbose: {}\n"
                   "volumes: {}\n"
                   .format(path, build_arg, cpu_shares,
                           config, context_file_limit, cpuset_cpus, file,
                           force_rm, label, memory, memory_swap,
                           no_cache, pull, quiet, rm,
                           tag, test_config, verbose,
                           volume))

        click.echo("loaded config:\n{}".format(str(config_dict)))
        return

    image = incubator.api.build(buildargs=build_arg,
                                path=path,
                                config=config_dict,
                                pull=pull,
                                dockerfile=file,
                                container_limits=limits,
                                rm=rm,
                                forcerm=force_rm,
                                volumes=volume,
                                tag=tag,
                                labels=label)

    output = str(image.id)
    if not quiet:
        output = "================================\n" \
                 "ID:{}\n" \
                 "================================\n" \
            .format(output)

    click.echo(output)


cli.add_command(build)

if __name__ == '__main__':
    cli()



"""
Class for managing configuration.
"""

import jsonschema
import six

from .constants import (CONFIG_BUILDARGS, CONFIG_CONTAINER_LIMITS,
                        CONFIG_CONTEXT_FILE_LIMIT, CONFIG_INFINITE_COMMAND,
                        CONFIG_LABELS, CONFIG_LAYERS, CONFIG_MKDIR_COMMAND,
                        CONFIG_PULL, CONFIG_TAGS, CONFIG_VERSION,
                        CONFIG_VOLUMES, DEFAULT_CONTEXT_FILE_LIMIT,
                        DEFAULT_TAG, DEFAULT_VERSION, INFINITE_COMMAND,
                        MKDIR_COMMAND)
from .utils import volumes_to_dict

schema_v1 = {
    "type": "object",
    "properties": {
        CONFIG_VERSION: {"type": "number"},
        CONFIG_LAYERS: {
            "type": "array",
            "items": {
                "type": "integer",
                "minimum": 1
            },
        },
        CONFIG_VOLUMES: {
            "properties": {},
            "patternProperties": {
                "": {
                    "type": "string"
                }
            }
        },
        CONFIG_LABELS: {
            "properties": {},
            "patternProperties": {
                "": {
                    "type": "string"
                }
            }
        },
        CONFIG_INFINITE_COMMAND: {"type": "string"},
        CONFIG_MKDIR_COMMAND: {'type': 'string'},
        CONFIG_CONTEXT_FILE_LIMIT: {'type': 'integer'},
        CONFIG_TAGS: {
            "type": "array",
            "items": {
                "type": "string"
            }
        }
    },
}

schema = {1.0: schema_v1}


class ImageConfig(object):
    """
    Representation of config file.
    Structure of file is described and validated by jsonschema.
    """

    def __init__(self, config_dict=None, pull=False,
                 rm=False, forcerm=False, buildargs=None,
                 container_limits=None, labels=None,
                 layers=None, tag=None,
                 secret_volumes=None,
                 infinite_command=None,
                 mkdir_command=None,
                 context_file_limit=None,
                 ):

        config_dict = config_dict or {}
        config_dict.setdefault(CONFIG_VERSION, DEFAULT_VERSION)
        self.__config = config_dict

        self._update_dict(CONFIG_BUILDARGS, buildargs)
        self._update_dict(CONFIG_CONTAINER_LIMITS, container_limits)
        self._update_dict(CONFIG_VOLUMES, volumes_to_dict(secret_volumes))
        self._update_dict(CONFIG_LABELS, labels)

        self._update_value(CONFIG_CONTEXT_FILE_LIMIT, context_file_limit, DEFAULT_CONTEXT_FILE_LIMIT)
        self._update_value(CONFIG_PULL, pull, False)

        self._update_sorted_array(CONFIG_LAYERS, layers)

        self._update_sorted_array(CONFIG_TAGS, tag)

        self._update_value(CONFIG_INFINITE_COMMAND, infinite_command, INFINITE_COMMAND)
        self._update_value(CONFIG_MKDIR_COMMAND, mkdir_command, MKDIR_COMMAND)

    def _update_dict(self, key, value):
        value = value or {}
        self.__config.setdefault(key, {})
        self.__config[key].update(value)

    def _update_sorted_array(self, key, value):
        value = value or []
        if isinstance(value, six.string_types):
            value = [value]
        self.__config.setdefault(key, [])
        self.__config[key] = list(set(self.__config[key] + value))
        self.__config[key].sort()

    def _update_value(self, key, value, default_value):
        self.__config.setdefault(key, default_value)
        if value is not None:
            self.__config[key] = value

    @property
    def config(self):
        """

        :return: config as a dictionary
        """
        return self.__config

    @config.setter
    def config(self, config):
        self.__config = config

    @property
    def layers(self):
        """

        :return: list with positions for commits
            (between which commands shout be new layer)
        """
        return self.__config.get(CONFIG_LAYERS, [])

    @layers.setter
    def layers(self, layers):
        self.__config[CONFIG_LAYERS] = layers

    @property
    def volumes(self):
        """

        :return: Tuples with source:destination volume only for build.
        """
        return self.__config.get(CONFIG_VOLUMES, [])

    @volumes.setter
    def volumes(self, volumes):
        self.__config[CONFIG_VOLUMES] = volumes

    @property
    def labels(self):
        """

        :return: Label tuples, which should be added to image.
        """
        return self.__config.get(CONFIG_LABELS, [])

    @labels.setter
    def labels(self, labels):
        self.__config[CONFIG_LABELS] = labels

    @property
    def container_limits(self):
        return self.__config.get(CONFIG_CONTAINER_LIMITS, {})

    @container_limits.setter
    def container_limits(self, limits):
        self.__config[CONFIG_CONTAINER_LIMITS] = limits

    @property
    def version(self):
        """

        :return: Version of structure from json.
        """
        return self.__config.get(CONFIG_VERSION, DEFAULT_VERSION)

    @property
    def infinite_command(self):
        return self.__config.get(CONFIG_INFINITE_COMMAND, INFINITE_COMMAND)

    @infinite_command.setter
    def infinite_command(self, command):
        self.__config[CONFIG_INFINITE_COMMAND] = command

    @property
    def mkdir_command(self):
        return self.__config.get(CONFIG_MKDIR_COMMAND, MKDIR_COMMAND)

    @mkdir_command.setter
    def mkdir_command(self, command):
        self.__config[CONFIG_MKDIR_COMMAND] = command

    @property
    def tags(self):
        return self.__config.get(CONFIG_TAGS, [])

    @tags.setter
    def tags(self, tags):
        self.__config[CONFIG_TAGS] = tags

    @property
    def tags_and_repos(self):
        """
        :return: list with tupple (repo,tag)
        """
        result = []
        for t in self.tags:
            t_split = t.split(':')
            t_part_number = len(t_split)
            if t_part_number > 2:
                raise Exception("Wrong tag format.")
            elif t_part_number == 2:
                result.append((t_split[0], t_split[1]))
            elif t_part_number == 1:
                result.append((t, DEFAULT_TAG))
        return result

    @property
    def pull(self):
        return self.__config.get(CONFIG_PULL, False)

    @pull.setter
    def pull(self, pull):
        self.__config[CONFIG_TAGS] = pull

    def is_valid(self):
        """

        :return: True if the json structure is valid, else False.
        """
        try:
            self.validate()
            return True
        except jsonschema.exceptions.ValidationError:
            return False

    @property
    def context_file_limit(self):
        return self.__config.get(CONFIG_CONTEXT_FILE_LIMIT, -1)

    def validate(self):
        """

        :return: None if the structure of json is valid, else throws exceptions.
        """
        version = self.__config.get(CONFIG_VERSION, None)
        version_schema = schema.get(version, None)

        if version_schema is not None:
            jsonschema.validate(self.__config, schema)
            return None
        else:
            raise jsonschema.exceptions.ValidationError("Unknown config version.")



"""
Core functionality of builder.
"""

import logging

import dockerfile_parse

from .commands import get_layers_from_dockerfile_structure
from .constants import APP_NAME
from .utils import ResultImage

logger = logging.getLogger(APP_NAME)


class Builder:
    """
    Class for building docker images.
    """

    def __init__(self, context, dockerfile, client, config, cache):

        self._dfp = dockerfile_parse.DockerfileParser(fileobj=dockerfile,
                                                      cache_content=True,
                                                      env_replace=False)

        self._context = context
        self._config = config
        self.client = client
        self._last_image = None
        self._cache = cache
        self._logs = []

    def build(self):
        """
        Builds an image from given Dockerfile, context and configuration.
        Using given container client.
        :return: id of final image
        """
        self._last_image = None
        self._logs = []
        self._prebuild()

        commands = self._dfp.structure
        layers = get_layers_from_dockerfile_structure(structure=commands,
                                                      layers_split_mark=self._config.layers,
                                                      build_context=self._context)

        for layer in layers:
            logger.info("================ LAYER {} ================\n".format(str(layer.number)))
            cached_image = self._cache.get_layer(self._last_image, layer)
            if cached_image:
                self._last_image = cached_image
            else:
                self._create_new_layer(layer)
                for cmd in layer.commands:
                    cmd.apply(self.client, self._config)

                if not layer.is_last:
                    self._middle_commit(layer)

        self._final_commit(layers[-1])

        self._postbuild()

        return ResultImage(image_id=self._last_image, logs=self._logs)

    def _create_base_container(self):
        """
        Creates a base container filled in Dockerfile.
        :returns id of base container
        """
        image = self._dfp.baseimage
        logger.debug("------> base image: {}".format(image))

        if self._config.pull or self.client.get_image(image) is None:
            self.client.pull_image(image)

        return self.client.run_container_infinitely(
            image=image,
            infinite_command=self._config.infinite_command,
            volumes=self._config.volumes,
            container_limits=self._config.container_limits)

    def _create_new_layer(self, layer):
        if not self._last_image:
            layer.conf.current_container = self._create_base_container()
        else:
            layer.conf.current_container = self.client.run_container_infinitely(
                image=self._last_image,
                infinite_command=self._config.infinite_command,
                volumes=self._config.volumes,
                container_limits=self._config.container_limits)

    def _final_commit(self, layer):
        self.client.stop_container(layer.conf.current_container)
        logger.debug("================ FINAL COMMIT ================\n")
        layer.conf.add_labels(self._config.labels)
        image = self.client.commit_container(layer.conf.current_container,
                                             conf=layer.conf.commit_config,
                                             author=layer.conf.author,
                                             message=layer.commit_message,
                                             tags=self._config.tags_and_repos)
        self._cache.add(image_id=image,
                        parent_id=self._last_image,
                        layer=layer)
        self._logs += layer.conf.logs
        self._last_image = image
        self.client.remove_container(layer.conf.current_container)

    def _middle_commit(self, layer):
        self.client.stop_container(layer.conf.current_container)
        logger.debug("================ COMMIT ================\n")
        image = self.client.commit_container(layer.conf.current_container,
                                             conf=layer.conf.commit_config,
                                             author=layer.conf.author,
                                             message=layer.commit_message)

        self._cache.add(image_id=image,
                        parent_id=self._last_image,
                        layer=layer)
        self._logs += layer.conf.logs
        self._last_image = image
        self.client.remove_container(layer.conf.current_container)

    def _prebuild(self):
        """
        For future use.
        """
        logger.debug('======> prebuild')

    def _postbuild(self):
        """
        For future use.
        """
        logger.debug('======> postbuild')
